/*
** Module   : AcquireMutex.java
** Abstract : directive to obtain ownership of a named mutex
**
** Copyright (c) 2009-2021, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description----------------------------------------
** 001 GES 20090224 Directive to obtain ownership of a named mutex.
** 002 CA  20210827 Refactored to allow other test types, beside terminal (which can be executed from the same 
**                  test suite).  Added SOAP and REST web service tests.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You should have received a copy of the GNU Affero General Public License
** along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.goldencode.harness.test;

import java.util.concurrent.locks.*;
import com.goldencode.harness.*;

/**
 * Directive to obtain ownership of a named mutex.
 */
public class AcquireMutex
extends Report
implements Testable
{
   /** Short name for the activity. */
   private static final String NAME = "ACQUIRE-MUTEX";
   
   /** The mutex to utilize. */
   private final ReentrantLock mutex;
   
   /**
    * Constructor.
    *
    * @param    name
    *           Mutex name.
    * @param    mutex
    *           The mutex semaphore to use.
    */
   public AcquireMutex(String name, ReentrantLock mutex)
   {
      super(NAME, String.format("name = '%s';", name));
      this.mutex = mutex;
   }
   
   /**
    * Execute the directive of the test using the given terminal as the data
    * sink and/or source for the operation.
    */
   @Override
   public void execute()
   {
      long start = System.currentTimeMillis();
      
      try
      {
         mutex.lockInterruptibly();
      }
      
      catch (InterruptedException ie)
      {
         long elapsed = System.currentTimeMillis() - start;
         setReason(String.format("Interrupted after %d milliseconds.", 
                                 elapsed));
         throw new RuntimeException(ie);
      }
      
      setStatus(Status.PASSED, null);
   }
}
