/*
** Module   : IndentingWriter.java
** Abstract : allows caller-controlled indenting of output 
**
** Copyright (c) 2008-2009, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20080309 Created initial version which allows caller-controlled
**                  indenting of output.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You should have received a copy of the GNU Affero General Public License
** along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.goldencode.io;

import java.io.*;

/**
 * Provides <code>Writer</code> semantics with caller-controlled
 * indenting.  An indent is written (inserted) between each newline and any
 * subsequent (non-newline) characters that are written.
 * <p>
 * This class is not thread-safe.
 *
 * @author    GES
 */
public class IndentingWriter
extends FilterWriter
implements Indentable
{
   /** Indent state manager. */
   private final Indentable indent = new IndentHandler();
   
   /** The last byte written to the stream. */
   private char last = '\n';
   
   /**
    * Creates an instance with the associated writer as the base and
    * a default 0 indent level.
    *
    * @param    out
    *           The writer on which we filter.
    */
   public IndentingWriter(Writer out)
   {
      super(out);
   }

   /**
    * Creates an instance with the associated writer as the base.
    *
    * @param    out
    *           The writer on which we filter.
    * @param    indent
    *           The initial indent level.
    */
   public IndentingWriter(Writer out, int indent)
   {
      this(out);
      this.indent.setIndent(indent);
   }

   /**
    * Accesses the current indent level.
    *
    * @return   Indent level.
    */
   public int getIndent()
   {
      return indent.getIndent();
   }

   /**
    * Sets the current indent level.
    *
    * @param    indent
    *           New indent level.
    */
   public void setIndent(int indent)
   {
      this.indent.setIndent(indent);
   }

   /**
    * Increments the indent level by 1.
    */
   public void increment()
   {
      indent.increment();
   }

   /**
    * Decrements the indent level by 1.
    */
   public void decrement()
   {
      indent.decrement();
   }

   /**
    * Accesses the number of spaces in a single indent level.
    *
    * @return   Width of each indent level.
    */
   public int getWidth()
   {
      return indent.getWidth();
   }

   /**
    * Sets the number of spaces in a single indent level.
    *
    * @param    width
    *           New indent width.
    */
   public void setWidth(int width)
   {
      indent.setWidth(width);
   }

   /**
    * Sets both the indent level and the number of spaces in a single indent
    * level. This is a convenience function to avoid 2 separate calls and
    * the small amount of shared overhead that would have been incurred in
    * those calls.
    *
    * @param    indent
    *           New indent level.
    * @param    width
    *           New indent width.
    */
   public void setPad(int indent, int width)
   {
      this.indent.setPad(indent, width);
   }
   
   /**
    * Get the current pad.
    *
    * @return   The pad string for the next line of output.
    */
   public String getPad()
   {
      return indent.getPad();
   }
   
   /**
    * Writes the specified character to this writer (only the low order 16
    * bits are written, the high order 16 bits are ignored).
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline characters.
    * 
    * @param    ch
    *           The character to write. 
    *
    * @throws   IOException
    *           if an I/O error occurs. In particular, an IOException may be
    *           thrown if the writer has been closed.
    */
   public void write(int ch)
   throws IOException
   {
      write(new char[] { (char) ch });
   }
      
   /**
    * Writes <code>len</code> characters from the specified string starting
    * at offset <code>off</code> in this writer. Element <code>off</code> is
    * the first character written and <code>off+len-1</code> is the last
    * character written.
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline characters.
    * 
    * @param    str
    *           The string to write. 
    * @param    off
    *           The start offset in the string.
    * @param    len
    *           The number of characters to write.
    *
    * @throws   IOException
    *           If an I/O error occurs. In particular, an IOException may be
    *           thrown if the writer has been closed.
    *
    * @throws   NullPointerException   
    *           If the string is <code>null</code>. 
    * @throws   IndexOutOfBoundsException   
    *           If the offset or length are negative, or if offset + length
    *           is larger than the length of the string.
    */
   public void write(String str, int off, int len)
   throws IOException,
          NullPointerException,
          IndexOutOfBoundsException
   {
      write(str.toCharArray(), off, len);
   }
   
   /**
    * Writes <code>len</code> characters from the specified character array
    * starting at offset <code>off</code> in this writer. Element ch[off] is
    * the first character written and ch[off+len-1] is the last character
    * written.
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline characters.
    * 
    * @param    ch
    *           The array of characters to write. 
    * @param    off
    *           The start offset in the array.
    * @param    len
    *           The number of characters to write.
    *
    * @throws   IOException
    *           If an I/O error occurs. In particular, an IOException may be
    *           thrown if the writer has been closed.
    *
    * @throws   NullPointerException   
    *           If the array is <code>null</code>. 
    * @throws   IndexOutOfBoundsException   
    *           If the offset or length are negative, or if offset + length
    *           is larger than the length of the array.
    */
   public void write(char[] ch, int off, int len)
   throws IOException,
          NullPointerException,
          IndexOutOfBoundsException
   {
      boolean empty = (indent.getIndent() == 0 && indent.getWidth() == 0);
      
      if (empty)
      {
         // no identing needed, use faster approach
         out.write(ch, off, len);
         
         // remember the last char
         last = ch[off + len - 1];
      }
      else
      {
         // indent needed, process one byte at a time
         for (int i = 0; i < len; i++)
         {
            char next = ch[i + off];
            
            // look for an output of a non-newline that follows a newline
            if (last == '\n' && next != '\n')
            {
               // insert the indent
               out.write(indent.getPad().toCharArray());
            }
            
            // write the current character
            out.write(next);
            
            // update the last character
            last = next;
         }
      }
   }
}
